import json
from datetime import date, datetime, timedelta

from shapely import Point

from cqlalchemy.stac import query as q
from cqlalchemy.stac.query import QueryBuilder, _DateTimeEncoder, filter_grouping

if __name__ == '__main__':
    import uuid
    query = q.QueryBuilder().datetime.equals(date.today()).\
        updated.lt(datetime.now()).\
        created.delta(date.today(), timedelta(45))
    query.sar.resolution_azimuth.lte(45)
    query.id.equals(str(uuid.uuid4()))
    query.geometry.intersects(Point(4, 5))
    query.filter(((query.sar.resolution_azimuth == 45.1) | (query.datetime == 55.1) | (query.id == "stuff.1")) |
                 ((query.created >= datetime.now()) | (query.created <= datetime.now())))
    query.filter(filter_grouping((query.sar.resolution_azimuth >= 90.2) | (query.id == "pancakes.2")) &
                 (query.sar.resolution_azimuth == 45.2) & (query.datetime == 55.2) & (query.id == "stuff.2") &
                 (query.created >= datetime.now()) & (query.created <= datetime.now()))
    query.filter((query.sar.resolution_azimuth == 45.3) & (query.datetime == 55.3) & (query.id == "stuff.3") &
                 (query.created >= datetime.now()) & (query.created <= datetime.now()) &
                 filter_grouping((query.sar.resolution_azimuth >= 90.3) | (query.id == "pancakes.3")))
    print(json.dumps(query.query_dump(), indent=4, cls=_DateTimeEncoder))

    print(QueryBuilder().sar.resolution_azimuth.gt(99).sar.resolution_azimuth.lt(1).query_dump())
    print(QueryBuilder().filter(query.sar.resolution_azimuth > 100))



import logging
import pkgutil
from string import Template

from cqlalchemy import __version__

logger = logging.getLogger(__name__)

enum_template = Template(pkgutil.get_data(__name__, "templates/enum.template").decode('utf-8'))
extension_template = Template(pkgutil.get_data(__name__, "templates/extension.template").decode('utf-8'))
query_template = Template(pkgutil.get_data(__name__, "templates/query.template").decode('utf-8'))
common_template = Template(pkgutil.get_data(__name__, "templates/common.template").decode('utf-8'))
common_docs_template = Template(pkgutil.get_data(__name__, "templates/common.docs.template").decode('utf-8'))
enum_custom_method_template = Template(pkgutil.get_data(__name__, "templates/enum.custom.method.template").decode('utf-8'))
number_query_attr_template = Template(pkgutil.get_data(__name__, "templates/number.query.attr.template").decode('utf-8'))
enum_query_attr_template = Template(pkgutil.get_data(__name__, "templates/enum.query.attr.template").decode('utf-8'))

ENUM_MEMBERS = "    {member} = \"{value}\"\n"
DATETIME_QUERY_EXT_ATTR = "        self.{partial_name} = _DateQuery(\"{field_name}\", query_block)\n"
BOOLEAN_QUERY_EXT_ATTR = "        self.{partial_name} = _BooleanQuery(\"{field_name}\", query_block)\n"
DATETIME_QUERY_ATTR = "        self.{partial_name} = _DateQuery(\"{field_name}\", self)\n"
STRING_QUERY_EXT_ATTR = "        self.{partial_name} = _StringQuery(\"{field_name}\", query_block)\n"
STRING_QUERY_ATTR = "        self.{partial_name} = _StringQuery(\"{field_name}\", self)\n"
GEOMETRY_QUERY_ATTR = "        self.{partial_name} = _SpatialQuery(\"{field_name}\", query_block)\n"
EXTENSION_ATTR = "\n        self.{jsond_prefix} = {class_name}(self)"
ATTR_DOC = "    {partial_name} : {class_name}\n        datetime query interface for searching items by the end_datetime field"


def build_enum(field_name: str, enum_object: dict, full_name=False, add_unique=False, field_description=""):
    prefix = ""
    if full_name and ":" in field_name:
        prefix = field_name.split(":")[0].upper()
    if ":" in field_name:
        field_name = field_name.split(":")[1]
    class_name = prefix + "".join([x.capitalize() for x in field_name.split("_")])

    member_definitions = ""
    custom_methods = ""
    for x in enum_object["enum"]:
        member = str(x).replace("-", "_").replace(" ", "_").strip()
        member_definitions += ENUM_MEMBERS.format(member=member, value=x)
        if add_unique:
            custom_methods += enum_custom_method_template.substitute(member=member, class_name=class_name) + "\n"

    field_description_name = " ".join([x.capitalize() for x in field_name.strip().split("_")])
    enum_query_description = f"\n    \"\"\"\n    {field_description_name} Enum Query Interface\n    \"\"\"\n"
    if field_description:
        enum_query_description = f"\n    \"\"\"\n    {field_description_name} Enum Query Interface\n    {field_description}\n    \"\"\"\n"
    enum_description = f"\n    \"\"\"\n    {field_description_name} Enum\n    \"\"\"\n"

    return enum_template.substitute(class_name=class_name,
                                    member_definitions=member_definitions,
                                    custom_methods=custom_methods,
                                    enum_query_description=enum_query_description,
                                    enum_description=enum_description), class_name


def other_attr_doc(partial_name, class_name, field_name):
    data_type = "enum"
    if class_name == "_DateQuery":
        data_type = "datetime"
    elif class_name == "_StringQuery":
        data_type = "string"
    elif class_name == "_SpatialQuery":
        data_type = "geometry"

    return f"    {partial_name} : {class_name}\n        {data_type} query interface for searching items by the {field_name} field\n"


def number_attr_doc(partial_name, field_name, min_value=None, max_value=None, is_int=False, additional_info=""):
    doc = f"    {partial_name}: _NumberQuery\n        number query interface for searching items by the {field_name} field"
    if min_value is not None:
        doc += f" where the minimum value is {min_value}"
        if max_value is not None:
            doc += f" and the max value is {max_value}"
    elif max_value is not None:
        doc += f" where the maximum value is {max_value}"
    doc += ". Float input."
    if is_int:
        doc += ". Integer input."
    if additional_info:
        doc += additional_info
    return doc + "\n"


class ExtensionBuilder:
    extension = ""
    class_name = ""
    jsond_prefix = ""

    def __init__(self, extension_schema, force_string_enum=False, fields_to_exclude=None, add_unique_enum=False):
        schema_url = extension_schema['$id']
        if fields_to_exclude is None:
            fields_to_exclude = []
        if "description" in extension_schema:
            extension_description = extension_schema["description"]
        elif "title" in extension_schema:
            # TODO 'https://stac-extensions.github.io/storage/v2.0.0/schema.json' is the only reason for this if statement
            extension_description = extension_schema["title"]
        else:
            logger.error(f"schema contains no description or title {schema_url}")
            return
        defs_key = "definitions"
        if defs_key not in extension_schema:
            if "$defs" not in extension_schema:
                # TODO 'https://stac-extensions.github.io/item-assets/v1.0.0/schema.json#', not sure how to handle this
                logger.error(f"contains no definitions field in schema {schema_url}")
                return
            # TODO 'https://stac-extensions.github.io/mlm/v1.3.0/schema.json' is the only reason for this if block
            defs_key = "$defs"
        definitions = extension_schema[defs_key]

        field_names = []
        if "v0" in extension_schema["$id"]:
            logger.error(f"skipping any extensions that are ~v0. This extension is {schema_url}")
            return
        if "fields" in definitions and "properties" in definitions["fields"]:
            field_names = [x for x in list(definitions["fields"]["properties"].keys()) if ":" in x]
            if len(field_names) != len(list(definitions["fields"]["properties"].keys())):
                logger.warning(f"{field_names} less than {list(definitions['fields'].keys())}")
            if len(field_names) > 0 and field_names[0] not in definitions:
                definitions = definitions["fields"]["properties"]
            elif len(field_names) == 0:
                # TODO https://stac-extensions.github.io/render/v1.0.0/schema.json (no json-ld)
                #   https://stac-extensions.github.io/themes/v1.0.0/schema.json (no json-ld)
                #   'https://stac-extensions.github.io/version/v1.2.0/schema.json#'
                logger.error(f"skipping any extensions that are ~v0. This extension is {schema_url}")
                return
            else:
                # TODO 'https://stac-extensions.github.io/mlm/v1.3.0/schema.json'
                #   'https://stac-extensions.github.io/pointcloud/v2.0.0/schema.json#'
                #   'https://stac-extensions.github.io/raster/v2.0.0/schema.json#'
                #   'https://stac-extensions.github.io/pointcloud/v2.0.0/schema.json#'
                logger.warning(f"odd formating for {schema_url}")
        elif "v1" in extension_schema["$id"]:
            # TODO 'https://stac-extensions.github.io/osc/v1.0.0-rc.3/schema.json#'
            #   https://stac-extensions.github.io/tiled-assets/v1.0.0/schema.json
            #   https://stac-extensions.github.io/usfws-nwi/v1.0.0/schema.json
            #   https://stac-extensions.github.io/web-map-links/v1.2.0/schema.json
            #   https://stac-extensions.github.io/authentication/v1.1.0/schema.json
            #   https://stac-extensions.github.io/ml-model/v1.0.0/schema.json
            #   'https://stac-extensions.github.io/alternate-assets/v1.2.0/schema.json#'
            #   'https://stac-extensions.github.io/language/v1.0.0/schema.json#'
            logger.error(f"skipping any extensions that are v1 and missing definitions[\"fields\"][\"properties\"]. This extension is {schema_url}")
            return

        if "fields" in definitions and "properties" not in definitions["fields"]:
            # TODO 'https://stac-extensions.github.io/landsat/v2.0.0/schema.json'
            field_names = [x for x in list(definitions["fields"].keys()) if ":" in x]
            if len(field_names) != len(list(definitions["fields"].keys())):
                logger.warning(f"{field_names} less than {list(definitions['fields'].keys())}")
        elif len(field_names) > 0 and field_names[0] not in definitions:
            # TODO 'https://stac-extensions.github.io/storage/v2.0.0/schema.json'
            logger.error(f"no fields discoverable for schema {schema_url}")

        if len(field_names) == 0:
            # TODO 'https://stac-extensions.github.io/storage/v2.0.0/schema.json'
            logger.error(f"no jsonld ':' separators found in keys {definitions}")
            return

        if field_names[0] not in definitions:
            # TODO 'https://stac-extensions.github.io/landsat/v2.0.0/schema.json'
            definitions = definitions["fields"]

        field_names.sort()
        self.jsond_prefix = field_names[0].split(":")[0]
        extension_name = self.jsond_prefix.capitalize()
        if extension_name not in extension_description:
            extension_name = extension_name.upper()
        self.class_name = f"_{extension_name}Extension"

        enum_definitions = ""
        attribute_instantiations = ""
        extension_attr_docs = ""
        for field_name in field_names:
            if field_name in fields_to_exclude:
                logger.info(f"skipping the field {field_name} as it is included in {fields_to_exclude}")
                continue
            partial_name = field_name.split(":")[1]
            field_obj = definitions[field_name]
            min_value = None
            if "minimum" in field_obj:
                min_value = field_obj["minimum"]
            max_value = None
            if "maximum" in field_obj:
                max_value = field_obj["maximum"]
            if "type" not in field_obj:
                # TODO 'https://stac-extensions.github.io/mlm/v1.3.0/schema.json'
                if "oneOf" in field_obj:
                    for o in field_obj["oneOf"]:
                        if "type" in o and o["type"] != "null":
                            field_obj.update(o)
                elif "anyOf" in field_obj:
                    for o in field_obj["anyOf"]:
                        if "type" in o and o["type"] != "null":
                            field_obj.update(o)
                if "type" not in field_obj:
                    if "$ref" in field_obj and "geojson" in field_obj["$ref"]:
                        field_obj["type"] = "geometry"
                    elif "$ref" in field_obj and field_obj["$ref"].startswith("#/"):
                        keys = field_obj["$ref"].strip("#").strip("/").split("/")
                        d = extension_schema
                        for k in keys:
                            if k in d:
                                d = d[k]
                            else:
                                break
                        if "type" in d:
                            field_obj.update(d)
                    if "type" not in field_obj:
                        logger.warning(f"{field_obj} does not contain type. skipping")
                        continue
            if "description" in field_obj:
                field_description = field_obj["description"]
            elif "title" in field_obj:
                field_description = field_obj["title"]
            else:
                field_description = ""

            if isinstance(field_obj["type"], list):
                if "null" in field_obj["type"]:
                    field_obj["type"].remove("null")
                elif len(field_obj["type"]) > 1:
                    logger.warning(f"array of {field_obj['type']} not supported as type")
                    continue
                if len(field_obj["type"]) == 1:
                    field_obj["type"] = field_obj["type"][0]
                else:
                    logger.warning(f"array of {field_obj['type']} not supported as type")
                    continue
            if field_obj["type"] == "number" or field_obj["type"] == "integer":
                is_int = field_obj["type"] == "integer"
                attribute_instantiations += number_query_attr_template.substitute(field_name=field_name,
                                                                                  partial_name=partial_name,
                                                                                  min_value=min_value,
                                                                                  max_value=max_value,
                                                                                  is_int=is_int)
                extension_attr_docs += number_attr_doc(partial_name=partial_name,
                                                       field_name=field_name,
                                                       min_value=min_value,
                                                       max_value=max_value,
                                                       is_int=is_int)
            elif field_obj["type"] == "string" and "format" in field_obj and field_obj["format"] == "date-time":
                attribute_instantiations += DATETIME_QUERY_EXT_ATTR.format(field_name=field_name,
                                                                           partial_name=partial_name)
                extension_attr_docs += other_attr_doc(partial_name=partial_name, class_name="_DateQuery", field_name=field_name)
            elif field_obj["type"] == "string" and "enum" in field_obj and not force_string_enum and not any(s[0].isdigit() for s in field_obj["enum"]):
                # landsat enum -> not any(s[0].isdigit() for s in field_obj["enum"])
                enum_definition, class_name = build_enum(field_name, field_obj, add_unique=add_unique_enum, field_description=field_description)
                enum_definitions += enum_definition
                enum_definitions += "\n\n"
                attribute_instantiations += enum_query_attr_template.substitute(field_name=field_name,
                                                                                partial_name=partial_name,
                                                                                class_name=class_name)
                e_class = f"_{class_name}Query"
                extension_attr_docs += other_attr_doc(partial_name=partial_name, class_name=e_class, field_name=field_name)
            elif field_obj["type"] == "string":
                attribute_instantiations += STRING_QUERY_EXT_ATTR.format(field_name=field_name,
                                                                         partial_name=partial_name)
                extension_attr_docs += other_attr_doc(partial_name=partial_name,
                                                      class_name="_StringQuery",
                                                      field_name=field_name)
            elif field_obj["type"] == "boolean":
                attribute_instantiations += BOOLEAN_QUERY_EXT_ATTR.format(field_name=field_name,
                                                                          partial_name=partial_name)
                extension_attr_docs += other_attr_doc(partial_name=partial_name,
                                                      class_name="_BooleanQuery",
                                                      field_name=field_name)
            elif field_obj["type"] == "geometry":
                attribute_instantiations += GEOMETRY_QUERY_ATTR.format(field_name=field_name,
                                                                       partial_name=partial_name)
                extension_attr_docs += other_attr_doc(partial_name=partial_name,
                                                      class_name="_SpatialQuery",
                                                      field_name=field_name)
            elif field_obj["type"] in ["array", "object"]:
                logger.info(f"not producing type {field_obj['type']}")
            else:
                raise ValueError(f"{field_obj['type']} not a processed type")

        extension_attr_docs = extension_attr_docs.rstrip()
        self.extension = enum_definitions + extension_template.substitute(class_name=self.class_name,
                                                                          extension_description=extension_description,
                                                                          extension_attr_docs=extension_attr_docs,
                                                                          attribute_instantiations=attribute_instantiations)


def build_query_file(extension_list: list[dict], fields_to_exclude=None, add_unique_enum=False) -> str:
    extension_definitions = ""
    extension_attributes = ""
    for extension_schema in extension_list:
        extension_builder = ExtensionBuilder(extension_schema, fields_to_exclude=fields_to_exclude, add_unique_enum=add_unique_enum)
        extension_definitions += f"\n\n{extension_builder.extension}"
        extension_attributes += EXTENSION_ATTR.format(jsond_prefix=extension_builder.jsond_prefix,
                                                      class_name=extension_builder.class_name)

    common_props_lines = common_template.substitute().split("\n")
    common_docs_props_lines = common_docs_template.substitute().split("\n")
    common_props = "\n"
    common_docs_props = "\n"
    if fields_to_exclude is not None:
        common_props_lines = [line for line in common_props_lines if not any(field in line for field in fields_to_exclude)]
        for i in range(0, len(common_props_lines)):
            line0 = common_docs_props_lines[i * 2]
            line1 = common_docs_props_lines[i * 2 + 1]
            if not any(field in line0 for field in fields_to_exclude):
                common_docs_props += line0 + "\n"
                common_docs_props += line1 + "\n"
    else:
        common_docs_props = "\n" + common_docs_template.substitute()

    common_props += "\n".join(common_props_lines)
    return query_template.substitute(cqlalchemy_version=__version__,
                                     extension_definitions=extension_definitions,
                                     common_attributes=common_props,
                                     common_docs=common_docs_props.rstrip(),
                                     extension_attributes=extension_attributes)



import json
import logging
import pathlib

import click
import requests

from cqlalchemy.scaffold.build import build_query_file

logger = logging.getLogger(__name__)


@click.command()
def build():
    """Create a cqlalchemy QueryBuilder class from STAC extensions."""
    extensions = {}
    stac_fields_to_ignore = set()

    click.echo("Enter extensions, either the path to a local file, a url or the extension json-ld name (sar, sat, etc):")
    while True:
        extension = click.prompt('STAC extension, file or url', default='', show_default=False)
        if extension == '':
            break
        try:
            if pathlib.Path(extension).exists():
                with pathlib.Path(extension).open() as fp:
                    extensions[extension] = json.load(fp)
            else:
                if not extension.startswith("http"):
                    logger.warning(f"treating input {extension} like extension json-ld code")
                    extension = f"https://raw.githubusercontent.com/stac-extensions/{extension}/refs/heads/main/json-schema/schema.json"
                response = requests.get(extension)
                response.raise_for_status()
                extensions[extension] = response.json()
        except BaseException as be:
            click.echo(f"{extension} failed with {be}")

    click.echo("Enter stac fields to omit from api or a path with a list of fields to omit:")
    while True:
        field = click.prompt('Field to ignore (or file of fields)', default='', show_default=False)
        if field == '':
            break
        try:
            if pathlib.Path(field).exists():
                with pathlib.Path(field).open("rt") as f:
                    for line in f.readlines():
                        stac_fields_to_ignore.add(line.strip())
            else:
                stac_fields_to_ignore.add(field)
        except BaseException as be:
            click.echo(f"{field} with {be}")

    add_unique_enum = click.prompt('Add unique enum fields for equals operator',
                                   default=False, type=bool, show_default=True)
    home_dir = pathlib.Path.home()
    default_file_location = home_dir / "query.py"
    if default_file_location.exists():
        output_file_location = click.prompt('Leave blank to overwrite. Or select a new location to save to',
                                            default=default_file_location, type=pathlib.Path, show_default=True)
    else:
        output_file_location = click.prompt('Define save location',
                                            default=default_file_location, type=pathlib.Path, show_default=True)

    # Step 3: Create an array using the sorted keys and their corresponding values
    sorted_extensions = [extensions[key] for key in sorted(list(extensions.keys()))]

    click.echo(f"Extensions: {[x['title'] for x in sorted_extensions if 'title' in x]}")
    click.echo(f"STAC fields to omit: {stac_fields_to_ignore}")
    query_file_data = build_query_file(sorted_extensions,
                                       fields_to_exclude=stac_fields_to_ignore,
                                       add_unique_enum=add_unique_enum)
    with pathlib.Path(output_file_location).open('wt') as f:
        f.write(query_file_data)


if __name__ == '__main__':
    build()



# # cqlalchemy/scaffold/extension.py
#
# import argparse
# from pyscaffold.extensions import Extension
# from pyscaffold.api import create_project
#
# class CQLAlchemyExtension(Extension):
#     """
#     This is a custom extension for PyScaffold that adds support for CQLAlchemy.
#
#     This extension provides additional functionality to generate libraries with multiple STAC extensions.
#     """
#
#     def augment_cli(self, parser):
#         parser.add_argument(
#             '--input-files',
#             help='Paths to the input text files to be parsed',
#             nargs='+',  # This allows multiple inputs
#             required=False
#         )
#         parser.add_argument(
#             '--interactive-mode',
#             help='Enable interactive mode',
#             action='store_true'
#         )
#
#     def activate(self, actions):
#         # Add your custom actions here
#         return self.register(actions, add_cqlalchemy, after="define_structure")
#
# def add_cqlalchemy(struct, opts):
#     if opts.get('interactive_mode'):
#         input_files = []
#         while True:
#             input_file = input("Enter the path to an input file (or press Enter to finish): ")
#             if not input_file:
#                 break
#             input_files.append(input_file)
#         opts['input_files'] = input_files
#
#     input_files = opts.get('input_files')
#     if input_files:
#         for input_file in input_files:
#             with open(input_file, 'r') as file:
#                 content = file.read()
#                 # Implement your custom logic to parse the file content
#                 print(f"Parsing content of {input_file}:")
#                 print(content)
#     return struct, opts

from typing import List

from pyscaffold import structure
from pyscaffold.actions import Action
from pyscaffold.extensions import Extension
from pyscaffold.operations import no_overwrite
from pyscaffold.templates import get_template

from cqlalchemy.scaffold import templates


class CQLAlchemyExtension(Extension):
    """
    This is a custom extension for PyScaffold that adds support for CQLAlchemy.
    """
    def augment_cli(self, parser):
        parser.add_argument(
            '--cqlactive',
            help='Enable interactive mode',
            action='store_true'
        )
        parser.add_argument(
            '--ignored-stac-fields-file',
            help='File with list of STAC item fields to not make accessible in query interface.',
            required=True
        )
        parser.add_argument(
            '--ignored-stac-fields',
            help='List of STAC item fields to not make accessible in query interface.',
            nargs='+',  # This allows multiple inputs
            required=False
        )
        parser.add_argument(
            '--extension-url-file',
            help='File with a list of STAC extension urls to be queried and parsed.',
            required=False
        )
        parser.add_argument(
            '--local-extension-files',
            help='Path(s) to the local STAC extension files to be parsed',
            nargs='+',  # This allows multiple inputs
            required=False
        )
        parser.add_argument(
            '--internal-stac-extensions',
            help='json-ld strings for internal STAC extensions',
            nargs='+',  # This allows multiple inputs
            required=False
        )

    def activate(self, actions: List[Action]) -> List[Action]:
        actions = self.register(actions, self.add_files)
        return self.register(actions, self.add_files)

    def add_files(self, struct, opts):
        pyproject_toml_template = get_template("pyproject.toml", relative_to=templates.__name__)
        files = {
            "pyproject.toml": (pyproject_toml_template, no_overwrite())
        }

        return structure.merge(struct, files), opts


def add_cqlalchemy(struct, opts):
    if opts.get('cqlactive'):
        input_files = []
        while True:
            input_file = input("Enter the path to an input file (or press Enter to finish): ")
            if not input_file:
                break
            input_files.append(input_file)
        opts['input_files'] = input_files
    return struct, opts



# This file is generated with version 0.0.5 of cqlalchemy https://github.com/davidraleigh/cqlalchemy

from __future__ import annotations

import json
import math
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from json import JSONEncoder

from shapely.geometry.base import BaseGeometry


class _DateTimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()


class _QueryTuple:
    def __init__(self, left, op: str, right):
        self.left = left
        self.op = op
        self.right = right

    def __or__(self, other):
        value = _QueryTuple(self, "|", other)
        if value.check_parents(bad_op="&"):
            raise ValueError("can't mix '&' and '|' in `filter` function. must be all 'or', "
                             "or all 'and' (except inside of the 'filter_grouping' function)")
        return value

    def __and__(self, other):
        value = _QueryTuple(self, "&", other)
        if value.check_parents(bad_op="|"):
            raise ValueError("can't mix '&' and '|' in `filter` function. must be all 'or', "
                             "or all 'and' (except inside of the 'filter_grouping' function)")
        return value

    def check_parents(self, bad_op):
        if isinstance(self.left, _QueryBase):
            return False

        if isinstance(self.left, _FilterTuple):
            if self.right.check_parents(bad_op):
                return True
            return False
        if isinstance(self.right, _FilterTuple):
            if self.left.check_parents(bad_op):
                return True
            return False

        if self.op == bad_op:
            return True
        if self.left.check_parents(bad_op):
            return True
        if self.right.check_parents(bad_op):
            return True
        return False

    @staticmethod
    def _recurse_build_query(query_tuple: _QueryTuple, filter_query: dict):
        if isinstance(query_tuple.left, _QueryBase):
            filter_query["args"].append({"op": query_tuple.op,
                                         "args": [query_tuple.left.property_obj, query_tuple.right]})
        elif isinstance(query_tuple.left, _FilterTuple):
            filter_query["args"].append(query_tuple.left._build_query())
            _QueryTuple._recurse_build_query(query_tuple.right, filter_query)
        elif isinstance(query_tuple.right, _FilterTuple):
            filter_query["args"].append(query_tuple.right._build_query())
            _QueryTuple._recurse_build_query(query_tuple.left, filter_query)
        else:
            _QueryTuple._recurse_build_query(query_tuple.left, filter_query)
            _QueryTuple._recurse_build_query(query_tuple.right, filter_query)
        return filter_query

    def _build_query(self):
        filter_query = {"op": "and", "args": []}
        filter_query = _QueryTuple._recurse_build_query(self, filter_query)
        if self.op == "|":
            filter_query["op"] = "or"
        return filter_query


class _FilterTuple(_QueryTuple):
    pass


class _QueryBase:
    def __init__(self, field_name, parent_obj: QueryBuilder):
        self._field_name = field_name
        self._parent_obj = parent_obj

    def _build_query(self):
        pass

    def __eq__(self, other):
        return _QueryTuple(self, "=", other)

    def __gt__(self, other):
        self._greater_check(other)
        return _QueryTuple(self, ">", other)

    def __ge__(self, other):
        self._greater_check(other)
        return _QueryTuple(self, ">=", other)

    def __lt__(self, other):
        self._less_check(other)
        return _QueryTuple(self, "<", other)

    def __le__(self, other):
        self._less_check(other)
        return _QueryTuple(self, "<=", other)

    @property
    def property_obj(self):
        return {"property": self._field_name}

    def _greater_check(self, value):
        pass

    def _less_check(self, value):
        pass

    def _check(self, value):
        pass


class _BooleanQuery(_QueryBase):
    _eq_value = None

    def equals(self, value: bool) -> QueryBuilder:
        """
        for the field, query for all items where it's boolean value equals this input

        Args:
            value (bool): equality check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._eq_value = value
        return self._parent_obj

    def _build_query(self):
        if self._eq_value is not None:
            return {
                "op": "=",
                "args": [self.property_obj, self._eq_value]
            }
        return None


class _BaseString(_QueryBase):
    _eq_value = None
    _in_values = None
    _like_value = None

    def _build_query(self):
        if self._eq_value is not None:
            return {
                "op": "=",
                "args": [self.property_obj, self._eq_value]
            }
        elif self._in_values is not None and len(self._in_values) > 0:
            return {
                "op": "in",
                "args": [
                    self.property_obj,
                    self._in_values
                ]
            }
        elif self._like_value is not None:
            return {
                "op": "like",
                "args": [
                    self.property_obj,
                    self._like_value
                ]
            }
        return None


class _EnumQuery(_BaseString):
    _enum_values: set[str] = set()

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        c = _EnumQuery(field_name, parent_obj)
        c._enum_values = set(enum_fields)
        if len(c._enum_values) <= 1:
            raise ValueError(f"enum_fields must have 2 or more unique values. fields are {enum_fields}")
        return c

    def _check(self, values: list[str]):
        if not set(values).issubset(self._enum_values):
            raise ValueError("")
        if self._in_values is not None or self._eq_value is not None or self._like_value is not None:
            raise ValueError("eq, in or like cannot already be set")


class _StringQuery(_BaseString):
    def equals(self, value: str) -> QueryBuilder:
        """
        for the field, query for all items where it's string value equals this input

        Args:
            value (str): equality check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._eq_value = value
        return self._parent_obj

    def in_set(self, values: list[str]) -> QueryBuilder:
        """
        for the values input, create an in_set query for this field

        Args:
            values (list[str]): for the values input, create an in_set query for this field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(values)
        self._in_values = values
        return self._parent_obj

    def like(self, value: str) -> QueryBuilder:
        """
        for the value input, create an like query for this field. Requires using the '%' operator within the value string for wildcard checking

        Args:
            value (str): for the value input, create an like query for this field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._like_value = value
        return self._parent_obj

    def _check(self, value):
        if self._in_values is not None or self._eq_value is not None or self._like_value is not None:
            raise ValueError("eq, in or like cannot already be set")


class _Query(_QueryBase):
    _gt_value = None
    _gt_operand = None
    _lt_value = None
    _lt_operand = None
    _eq_value = None

    def _build_query(self):
        if self._eq_value is not None:
            return {
                "op": "=",
                "args": [self.property_obj, self._eq_value]
            }
        elif self._gt_value is None and self._lt_value is None:
            return None

        gt_query = {
            "op": self._gt_operand,
            "args": [self.property_obj, self._gt_value]
        }
        lt_query = {
            "op": self._lt_operand,
            "args": [self.property_obj, self._lt_value]
        }
        if self._gt_value is not None and self._lt_value is None:
            return gt_query
        elif self._lt_value is not None and self._gt_value is None:
            return lt_query
        elif self._gt_value is not None and self._lt_value is not None and self._gt_value < self._lt_value:
            return {
                "op": "and",
                "args": [
                    gt_query, lt_query
                ]
            }
        elif self._gt_value is not None and self._lt_value is not None and self._gt_value > self._lt_value:
            return {
                "op": "or",
                "args": [
                    gt_query, lt_query
                ]
            }

    def equals(self, value) -> QueryBuilder:
        """
        for the field, query for all items where it's value equals this input

        Args:
            value: equality check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._eq_value = value
        return self._parent_obj

    def gt(self, value) -> QueryBuilder:
        """
        for the field, query for all items where it's value is greater than this input

        Args:
            value: value for greater than check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._greater_check(value)
        self._gt_value = value
        self._gt_operand = ">"
        return self._parent_obj

    def gte(self, value) -> QueryBuilder:
        """
        for the field, query for all items where it's value is greater than or equal to this input

        Args:
            value: value for greater than or equal to check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._greater_check(value)
        self._gt_value = value
        self._gt_operand = ">="
        return self._parent_obj

    def lt(self, value) -> QueryBuilder:
        """
        for the field, query for all items where it's value is less than this input

        Args:
            value: value for less than check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._less_check(value)
        self._lt_value = value
        self._lt_operand = "<"
        return self._parent_obj

    def lte(self, value) -> QueryBuilder:
        """
        for the field, query for all items where it's value is less than or equal to this input

        Args:
            value: value for less than or equal to check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        self._less_check(value)
        self._lt_value = value
        self._lt_operand = "<="
        return self._parent_obj


class _DateQuery(_Query):
    def equals(self, value: date, tzinfo=timezone.utc) -> QueryBuilder:
        """
        for the field, query for all items where it's date equals this input

        Args:
            value: equality check for the field.

        Returns:
            QueryBuilder: query builder for additional queries to add
        """
        self._check(value)
        if isinstance(value, datetime):
            self._eq_value = value
        elif isinstance(value, date):
            start = datetime.combine(value, datetime.min.time(), tzinfo=tzinfo)
            end = datetime.combine(value, datetime.max.time(), tzinfo=tzinfo)
            self._gt_value = start
            self._gt_operand = ">="
            self._lt_value = end
            self._lt_operand = "<="
        else:
            self._eq_value = value

        return self._parent_obj

    def delta(self, value: date, td: timedelta, tzinfo=timezone.utc):
        # self._equals_check()
        if td.total_seconds() > 0:
            start = datetime.combine(value, datetime.min.time(), tzinfo=tzinfo)
            end = start + td
        else:
            end = datetime.combine(value, datetime.max.time(), tzinfo=tzinfo)
            start = end + td
        self._gt_value = start
        self._gt_operand = ">="
        self._lt_value = end
        self._lt_operand = "<="
        return self._parent_obj

    def _check(self, value):
        if isinstance(value, datetime):
            if value.tzinfo is None:
                raise ValueError(f"datetime {value} does not have timezone set.")


class _NumberQuery(_Query):
    _min_value = None
    _max_value = None
    _is_int = False

    def equals(self, value):
        return super().equals(value)

    @classmethod
    def init_with_limits(cls, field_name, parent_obj: QueryBuilder, min_value=None, max_value=None, is_int=False):
        c = _NumberQuery(field_name, parent_obj)
        c._min_value = min_value
        c._max_value = max_value
        c._is_int = is_int
        return c

    def _greater_check(self, value):
        super(_NumberQuery, self)._greater_check(value)
        self._check_range(value)

    def _less_check(self, value):
        super(_NumberQuery, self)._less_check(value)
        self._check_range(value)

    def _check_range(self, value):
        if self._min_value is not None and value < self._min_value:
            raise ValueError(f"setting value of {value}, "
                             f"can't be less than min value of {self._min_value} for {self._field_name}")
        if self._max_value is not None and value > self._max_value:
            raise ValueError(f"setting value of {value}, "
                             f"can't be greater than max value of {self._max_value} for {self._field_name}")

    def _check(self, value):
        if self._is_int and not isinstance(value, int) and math.floor(value) != value:
            raise ValueError(f"for integer type, must use ints. {value} is not an int")


class _SpatialQuery(_QueryBase):
    _geometry = None

    def intersects(self, geometry: BaseGeometry) -> QueryBuilder:
        self._geometry = geometry
        return self._parent_obj

    def _build_query(self):
        if self._geometry is None:
            return None

        return {
            "op": "s_intersects",
            "args": [
                self.property_obj,
                self._geometry.__geo_interface__
            ]
        }


class _Extension:
    def __init__(self, query_block: QueryBuilder):
        self._filter_expressions: list[_QueryTuple] = []

    def _build_query(self):
        properties = list(vars(self).values())
        args = [x._build_query() for x in properties if isinstance(x, _QueryBase) and x._build_query() is not None]
        for query_filter in self._filter_expressions:
            args.append(query_filter._build_query())

        if len(args) == 0:
            return []
        return args


class CommonName(str, Enum):
    """
    Common Name Enum
    """

    pan = "pan"
    coastal = "coastal"
    blue = "blue"
    green = "green"
    green05 = "green05"
    yellow = "yellow"
    red = "red"
    rededge = "rededge"
    rededge071 = "rededge071"
    rededge075 = "rededge075"
    rededge078 = "rededge078"
    nir = "nir"
    nir08 = "nir08"
    nir09 = "nir09"
    cirrus = "cirrus"
    swir16 = "swir16"
    swir22 = "swir22"
    lwir = "lwir"
    lwir11 = "lwir11"
    lwir12 = "lwir12"


class _CommonNameQuery(_EnumQuery):
    """
    Common Name Enum Query Interface
    Common Name of the band
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _CommonNameQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: CommonName) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[CommonName]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def pan(self) -> QueryBuilder:
        return self.equals(CommonName.pan)

    def coastal(self) -> QueryBuilder:
        return self.equals(CommonName.coastal)

    def blue(self) -> QueryBuilder:
        return self.equals(CommonName.blue)

    def green(self) -> QueryBuilder:
        return self.equals(CommonName.green)

    def green05(self) -> QueryBuilder:
        return self.equals(CommonName.green05)

    def yellow(self) -> QueryBuilder:
        return self.equals(CommonName.yellow)

    def red(self) -> QueryBuilder:
        return self.equals(CommonName.red)

    def rededge(self) -> QueryBuilder:
        return self.equals(CommonName.rededge)

    def rededge071(self) -> QueryBuilder:
        return self.equals(CommonName.rededge071)

    def rededge075(self) -> QueryBuilder:
        return self.equals(CommonName.rededge075)

    def rededge078(self) -> QueryBuilder:
        return self.equals(CommonName.rededge078)

    def nir(self) -> QueryBuilder:
        return self.equals(CommonName.nir)

    def nir08(self) -> QueryBuilder:
        return self.equals(CommonName.nir08)

    def nir09(self) -> QueryBuilder:
        return self.equals(CommonName.nir09)

    def cirrus(self) -> QueryBuilder:
        return self.equals(CommonName.cirrus)

    def swir16(self) -> QueryBuilder:
        return self.equals(CommonName.swir16)

    def swir22(self) -> QueryBuilder:
        return self.equals(CommonName.swir22)

    def lwir(self) -> QueryBuilder:
        return self.equals(CommonName.lwir)

    def lwir11(self) -> QueryBuilder:
        return self.equals(CommonName.lwir11)

    def lwir12(self) -> QueryBuilder:
        return self.equals(CommonName.lwir12)


class _EOExtension(_Extension):
    """
    STAC EO Extension for STAC Items and STAC Collections.

    ...

    Attributes
    ----------
    center_wavelength: _NumberQuery
        number query interface for searching items by the eo:center_wavelength field. Float input.
    cloud_cover: _NumberQuery
        number query interface for searching items by the eo:cloud_cover field where the minimum value is 0 and the max value is 100. Float input.
    common_name : _CommonNameQuery
        enum query interface for searching items by the eo:common_name field
    full_width_half_max: _NumberQuery
        number query interface for searching items by the eo:full_width_half_max field. Float input.
    snow_cover: _NumberQuery
        number query interface for searching items by the eo:snow_cover field where the minimum value is 0 and the max value is 100. Float input.
    solar_illumination: _NumberQuery
        number query interface for searching items by the eo:solar_illumination field where the minimum value is 0. Float input.
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.center_wavelength = _NumberQuery.init_with_limits("eo:center_wavelength", query_block, min_value=None, max_value=None, is_int=False)
        self.cloud_cover = _NumberQuery.init_with_limits("eo:cloud_cover", query_block, min_value=0, max_value=100, is_int=False)
        self.common_name = _CommonNameQuery.init_enums("eo:common_name", query_block, [x.value for x in CommonName])
        self.full_width_half_max = _NumberQuery.init_with_limits("eo:full_width_half_max", query_block, min_value=None, max_value=None, is_int=False)
        self.snow_cover = _NumberQuery.init_with_limits("eo:snow_cover", query_block, min_value=0, max_value=100, is_int=False)
        self.solar_illumination = _NumberQuery.init_with_limits("eo:solar_illumination", query_block, min_value=0, max_value=None, is_int=False)


class CollectionCategory(str, Enum):
    """
    Collection Category Enum
    """

    A1 = "A1"
    A2 = "A2"
    T1 = "T1"
    T2 = "T2"
    RT = "RT"


class _CollectionCategoryQuery(_EnumQuery):
    """
    Collection Category Enum Query Interface
    Collection Category
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _CollectionCategoryQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: CollectionCategory) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[CollectionCategory]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def A1(self) -> QueryBuilder:
        return self.equals(CollectionCategory.A1)

    def A2(self) -> QueryBuilder:
        return self.equals(CollectionCategory.A2)

    def T1(self) -> QueryBuilder:
        return self.equals(CollectionCategory.T1)

    def T2(self) -> QueryBuilder:
        return self.equals(CollectionCategory.T2)

    def RT(self) -> QueryBuilder:
        return self.equals(CollectionCategory.RT)


class Correction(str, Enum):
    """
    Correction Enum
    """

    L1TP = "L1TP"
    L1GT = "L1GT"
    L1GS = "L1GS"
    L2SR = "L2SR"
    L2SP = "L2SP"


class _CorrectionQuery(_EnumQuery):
    """
    Correction Enum Query Interface
    Product Correction Level
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _CorrectionQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: Correction) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[Correction]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def L1TP(self) -> QueryBuilder:
        return self.equals(Correction.L1TP)

    def L1GT(self) -> QueryBuilder:
        return self.equals(Correction.L1GT)

    def L1GS(self) -> QueryBuilder:
        return self.equals(Correction.L1GS)

    def L2SR(self) -> QueryBuilder:
        return self.equals(Correction.L2SR)

    def L2SP(self) -> QueryBuilder:
        return self.equals(Correction.L2SP)


class _LandsatExtension(_Extension):
    """
    Landsat Extension to STAC Items.

    ...

    Attributes
    ----------
    cloud_cover_land: _NumberQuery
        number query interface for searching items by the landsat:cloud_cover_land field where the minimum value is -1 and the max value is 100. Float input.
    collection_category : _CollectionCategoryQuery
        enum query interface for searching items by the landsat:collection_category field
    collection_number : _StringQuery
        string query interface for searching items by the landsat:collection_number field
    correction : _CorrectionQuery
        enum query interface for searching items by the landsat:correction field
    product_generated : _DateQuery
        datetime query interface for searching items by the landsat:product_generated field
    scene_id : _StringQuery
        string query interface for searching items by the landsat:scene_id field
    wrs_path : _StringQuery
        string query interface for searching items by the landsat:wrs_path field
    wrs_row : _StringQuery
        string query interface for searching items by the landsat:wrs_row field
    wrs_type : _StringQuery
        string query interface for searching items by the landsat:wrs_type field
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.cloud_cover_land = _NumberQuery.init_with_limits("landsat:cloud_cover_land", query_block, min_value=-1, max_value=100, is_int=False)
        self.collection_category = _CollectionCategoryQuery.init_enums("landsat:collection_category", query_block, [x.value for x in CollectionCategory])
        self.collection_number = _StringQuery("landsat:collection_number", query_block)
        self.correction = _CorrectionQuery.init_enums("landsat:correction", query_block, [x.value for x in Correction])
        self.product_generated = _DateQuery("landsat:product_generated", query_block)
        self.scene_id = _StringQuery("landsat:scene_id", query_block)
        self.wrs_path = _StringQuery("landsat:wrs_path", query_block)
        self.wrs_row = _StringQuery("landsat:wrs_row", query_block)
        self.wrs_type = _StringQuery("landsat:wrs_type", query_block)


class Accelerator(str, Enum):
    """
    Accelerator Enum
    """

    amd64 = "amd64"
    cuda = "cuda"
    xla = "xla"
    amd_rocm = "amd-rocm"
    intel_ipex_cpu = "intel-ipex-cpu"
    intel_ipex_gpu = "intel-ipex-gpu"
    macos_arm = "macos-arm"


class _AcceleratorQuery(_EnumQuery):
    """
    Accelerator Enum Query Interface
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _AcceleratorQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: Accelerator) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[Accelerator]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def amd64(self) -> QueryBuilder:
        return self.equals(Accelerator.amd64)

    def cuda(self) -> QueryBuilder:
        return self.equals(Accelerator.cuda)

    def xla(self) -> QueryBuilder:
        return self.equals(Accelerator.xla)

    def amd_rocm(self) -> QueryBuilder:
        return self.equals(Accelerator.amd_rocm)

    def intel_ipex_cpu(self) -> QueryBuilder:
        return self.equals(Accelerator.intel_ipex_cpu)

    def intel_ipex_gpu(self) -> QueryBuilder:
        return self.equals(Accelerator.intel_ipex_gpu)

    def macos_arm(self) -> QueryBuilder:
        return self.equals(Accelerator.macos_arm)


class Framework(str, Enum):
    """
    Framework Enum
    """

    PyTorch = "PyTorch"
    TensorFlow = "TensorFlow"
    scikit_learn = "scikit-learn"
    Hugging_Face = "Hugging Face"
    Keras = "Keras"
    ONNX = "ONNX"
    rgee = "rgee"
    spatialRF = "spatialRF"
    JAX = "JAX"
    MXNet = "MXNet"
    Caffe = "Caffe"
    PyMC = "PyMC"
    Weka = "Weka"


class _FrameworkQuery(_EnumQuery):
    """
    Framework Enum Query Interface
    Any other framework name to allow extension. Enum names should be preferred when possible to allow better portability.
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _FrameworkQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: Framework) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[Framework]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def PyTorch(self) -> QueryBuilder:
        return self.equals(Framework.PyTorch)

    def TensorFlow(self) -> QueryBuilder:
        return self.equals(Framework.TensorFlow)

    def scikit_learn(self) -> QueryBuilder:
        return self.equals(Framework.scikit_learn)

    def Hugging_Face(self) -> QueryBuilder:
        return self.equals(Framework.Hugging_Face)

    def Keras(self) -> QueryBuilder:
        return self.equals(Framework.Keras)

    def ONNX(self) -> QueryBuilder:
        return self.equals(Framework.ONNX)

    def rgee(self) -> QueryBuilder:
        return self.equals(Framework.rgee)

    def spatialRF(self) -> QueryBuilder:
        return self.equals(Framework.spatialRF)

    def JAX(self) -> QueryBuilder:
        return self.equals(Framework.JAX)

    def MXNet(self) -> QueryBuilder:
        return self.equals(Framework.MXNet)

    def Caffe(self) -> QueryBuilder:
        return self.equals(Framework.Caffe)

    def PyMC(self) -> QueryBuilder:
        return self.equals(Framework.PyMC)

    def Weka(self) -> QueryBuilder:
        return self.equals(Framework.Weka)


class _MLMExtension(_Extension):
    """
    This object represents the metadata for a Machine Learning Model (MLM) used in STAC documents.

    ...

    Attributes
    ----------
    accelerator : _AcceleratorQuery
        enum query interface for searching items by the mlm:accelerator field
    accelerator_constrained : _BooleanQuery
        enum query interface for searching items by the mlm:accelerator_constrained field
    accelerator_count: _NumberQuery
        number query interface for searching items by the mlm:accelerator_count field where the minimum value is 1. Float input.. Integer input.
    accelerator_summary : _StringQuery
        string query interface for searching items by the mlm:accelerator_summary field
    architecture : _StringQuery
        string query interface for searching items by the mlm:architecture field
    batch_size_suggestion: _NumberQuery
        number query interface for searching items by the mlm:batch_size_suggestion field where the minimum value is 0. Float input.. Integer input.
    framework : _FrameworkQuery
        enum query interface for searching items by the mlm:framework field
    framework_version : _StringQuery
        string query interface for searching items by the mlm:framework_version field
    memory_size: _NumberQuery
        number query interface for searching items by the mlm:memory_size field where the minimum value is 0. Float input.. Integer input.
    name : _StringQuery
        string query interface for searching items by the mlm:name field
    pretrained : _BooleanQuery
        enum query interface for searching items by the mlm:pretrained field
    pretrained_source : _StringQuery
        string query interface for searching items by the mlm:pretrained_source field
    total_parameters: _NumberQuery
        number query interface for searching items by the mlm:total_parameters field where the minimum value is 0. Float input.. Integer input.
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.accelerator = _AcceleratorQuery.init_enums("mlm:accelerator", query_block, [x.value for x in Accelerator])
        self.accelerator_constrained = _BooleanQuery("mlm:accelerator_constrained", query_block)
        self.accelerator_count = _NumberQuery.init_with_limits("mlm:accelerator_count", query_block, min_value=1, max_value=None, is_int=True)
        self.accelerator_summary = _StringQuery("mlm:accelerator_summary", query_block)
        self.architecture = _StringQuery("mlm:architecture", query_block)
        self.batch_size_suggestion = _NumberQuery.init_with_limits("mlm:batch_size_suggestion", query_block, min_value=0, max_value=None, is_int=True)
        self.framework = _FrameworkQuery.init_enums("mlm:framework", query_block, [x.value for x in Framework])
        self.framework_version = _StringQuery("mlm:framework_version", query_block)
        self.memory_size = _NumberQuery.init_with_limits("mlm:memory_size", query_block, min_value=0, max_value=None, is_int=True)
        self.name = _StringQuery("mlm:name", query_block)
        self.pretrained = _BooleanQuery("mlm:pretrained", query_block)
        self.pretrained_source = _StringQuery("mlm:pretrained_source", query_block)
        self.total_parameters = _NumberQuery.init_with_limits("mlm:total_parameters", query_block, min_value=0, max_value=None, is_int=True)


class _ProjExtension(_Extension):
    """
    STAC Projection Extension for STAC Items.

    ...

    Attributes
    ----------
    code : _StringQuery
        string query interface for searching items by the proj:code field
    geometry : _SpatialQuery
        geometry query interface for searching items by the proj:geometry field
    wkt2 : _StringQuery
        string query interface for searching items by the proj:wkt2 field
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.code = _StringQuery("proj:code", query_block)
        self.geometry = _SpatialQuery("proj:geometry", query_block)
        self.wkt2 = _StringQuery("proj:wkt2", query_block)


class FrequencyBand(str, Enum):
    """
    Frequency Band Enum
    """

    P = "P"
    L = "L"
    S = "S"
    C = "C"
    X = "X"
    Ku = "Ku"
    K = "K"
    Ka = "Ka"


class _FrequencyBandQuery(_EnumQuery):
    """
    Frequency Band Enum Query Interface
    Frequency Band
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _FrequencyBandQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: FrequencyBand) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[FrequencyBand]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def P(self) -> QueryBuilder:
        return self.equals(FrequencyBand.P)

    def L(self) -> QueryBuilder:
        return self.equals(FrequencyBand.L)

    def S(self) -> QueryBuilder:
        return self.equals(FrequencyBand.S)

    def C(self) -> QueryBuilder:
        return self.equals(FrequencyBand.C)

    def X(self) -> QueryBuilder:
        return self.equals(FrequencyBand.X)

    def Ku(self) -> QueryBuilder:
        return self.equals(FrequencyBand.Ku)

    def K(self) -> QueryBuilder:
        return self.equals(FrequencyBand.K)

    def Ka(self) -> QueryBuilder:
        return self.equals(FrequencyBand.Ka)


class ObservationDirection(str, Enum):
    """
    Observation Direction Enum
    """

    left = "left"
    right = "right"


class _ObservationDirectionQuery(_EnumQuery):
    """
    Observation Direction Enum Query Interface
    Antenna pointing direction
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _ObservationDirectionQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: ObservationDirection) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[ObservationDirection]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def left(self) -> QueryBuilder:
        return self.equals(ObservationDirection.left)

    def right(self) -> QueryBuilder:
        return self.equals(ObservationDirection.right)


class _SARExtension(_Extension):
    """
    STAC SAR Extension to a STAC Item

    ...

    Attributes
    ----------
    center_frequency: _NumberQuery
        number query interface for searching items by the sar:center_frequency field. Float input.
    frequency_band : _FrequencyBandQuery
        enum query interface for searching items by the sar:frequency_band field
    instrument_mode : _StringQuery
        string query interface for searching items by the sar:instrument_mode field
    looks_azimuth: _NumberQuery
        number query interface for searching items by the sar:looks_azimuth field where the minimum value is 0. Float input.. Integer input.
    looks_equivalent_number: _NumberQuery
        number query interface for searching items by the sar:looks_equivalent_number field where the minimum value is 0. Float input.
    looks_range: _NumberQuery
        number query interface for searching items by the sar:looks_range field where the minimum value is 0. Float input.. Integer input.
    observation_direction : _ObservationDirectionQuery
        enum query interface for searching items by the sar:observation_direction field
    pixel_spacing_azimuth: _NumberQuery
        number query interface for searching items by the sar:pixel_spacing_azimuth field where the minimum value is 0. Float input.
    pixel_spacing_range: _NumberQuery
        number query interface for searching items by the sar:pixel_spacing_range field where the minimum value is 0. Float input.
    product_type : _StringQuery
        string query interface for searching items by the sar:product_type field
    resolution_azimuth: _NumberQuery
        number query interface for searching items by the sar:resolution_azimuth field where the minimum value is 0. Float input.
    resolution_range: _NumberQuery
        number query interface for searching items by the sar:resolution_range field where the minimum value is 0. Float input.
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.center_frequency = _NumberQuery.init_with_limits("sar:center_frequency", query_block, min_value=None, max_value=None, is_int=False)
        self.frequency_band = _FrequencyBandQuery.init_enums("sar:frequency_band", query_block, [x.value for x in FrequencyBand])
        self.instrument_mode = _StringQuery("sar:instrument_mode", query_block)
        self.looks_azimuth = _NumberQuery.init_with_limits("sar:looks_azimuth", query_block, min_value=0, max_value=None, is_int=True)
        self.looks_equivalent_number = _NumberQuery.init_with_limits("sar:looks_equivalent_number", query_block, min_value=0, max_value=None, is_int=False)
        self.looks_range = _NumberQuery.init_with_limits("sar:looks_range", query_block, min_value=0, max_value=None, is_int=True)
        self.observation_direction = _ObservationDirectionQuery.init_enums("sar:observation_direction", query_block, [x.value for x in ObservationDirection])
        self.pixel_spacing_azimuth = _NumberQuery.init_with_limits("sar:pixel_spacing_azimuth", query_block, min_value=0, max_value=None, is_int=False)
        self.pixel_spacing_range = _NumberQuery.init_with_limits("sar:pixel_spacing_range", query_block, min_value=0, max_value=None, is_int=False)
        self.product_type = _StringQuery("sar:product_type", query_block)
        self.resolution_azimuth = _NumberQuery.init_with_limits("sar:resolution_azimuth", query_block, min_value=0, max_value=None, is_int=False)
        self.resolution_range = _NumberQuery.init_with_limits("sar:resolution_range", query_block, min_value=0, max_value=None, is_int=False)


class OrbitState(str, Enum):
    """
    Orbit State Enum
    """

    ascending = "ascending"
    descending = "descending"
    geostationary = "geostationary"


class _OrbitStateQuery(_EnumQuery):
    """
    Orbit State Enum Query Interface
    Orbit State
    """

    @classmethod
    def init_enums(cls, field_name, parent_obj: QueryBuilder, enum_fields: list[str]):
        o = _OrbitStateQuery(field_name, parent_obj)
        o._enum_values = set(enum_fields)
        return o

    def equals(self, value: OrbitState) -> QueryBuilder:
        self._check([value.value])
        self._eq_value = value.value
        return self._parent_obj

    def in_set(self, values: list[OrbitState]) -> QueryBuilder:
        extracted = [x.value for x in values]
        self._check(extracted)
        self._in_values = extracted
        return self._parent_obj

    def ascending(self) -> QueryBuilder:
        return self.equals(OrbitState.ascending)

    def descending(self) -> QueryBuilder:
        return self.equals(OrbitState.descending)

    def geostationary(self) -> QueryBuilder:
        return self.equals(OrbitState.geostationary)


class _SatExtension(_Extension):
    """
    STAC Sat Extension to a STAC Item.

    ...

    Attributes
    ----------
    absolute_orbit: _NumberQuery
        number query interface for searching items by the sat:absolute_orbit field where the minimum value is 1. Float input.. Integer input.
    anx_datetime : _DateQuery
        datetime query interface for searching items by the sat:anx_datetime field
    orbit_cycle: _NumberQuery
        number query interface for searching items by the sat:orbit_cycle field where the minimum value is 1. Float input.. Integer input.
    orbit_state : _OrbitStateQuery
        enum query interface for searching items by the sat:orbit_state field
    platform_international_designator : _StringQuery
        string query interface for searching items by the sat:platform_international_designator field
    relative_orbit: _NumberQuery
        number query interface for searching items by the sat:relative_orbit field where the minimum value is 1. Float input.. Integer input.
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.absolute_orbit = _NumberQuery.init_with_limits("sat:absolute_orbit", query_block, min_value=1, max_value=None, is_int=True)
        self.anx_datetime = _DateQuery("sat:anx_datetime", query_block)
        self.orbit_cycle = _NumberQuery.init_with_limits("sat:orbit_cycle", query_block, min_value=1, max_value=None, is_int=True)
        self.orbit_state = _OrbitStateQuery.init_enums("sat:orbit_state", query_block, [x.value for x in OrbitState])
        self.platform_international_designator = _StringQuery("sat:platform_international_designator", query_block)
        self.relative_orbit = _NumberQuery.init_with_limits("sat:relative_orbit", query_block, min_value=1, max_value=None, is_int=True)


class _ViewExtension(_Extension):
    """
    STAC View Geometry Extension for STAC Items and STAC Collections.

    ...

    Attributes
    ----------
    azimuth: _NumberQuery
        number query interface for searching items by the view:azimuth field where the minimum value is 0 and the max value is 360. Float input.
    incidence_angle: _NumberQuery
        number query interface for searching items by the view:incidence_angle field where the minimum value is 0 and the max value is 90. Float input.
    off_nadir: _NumberQuery
        number query interface for searching items by the view:off_nadir field where the minimum value is 0 and the max value is 90. Float input.
    sun_azimuth: _NumberQuery
        number query interface for searching items by the view:sun_azimuth field where the minimum value is 0 and the max value is 360. Float input.
    sun_elevation: _NumberQuery
        number query interface for searching items by the view:sun_elevation field where the minimum value is -90 and the max value is 90. Float input.
    """
    def __init__(self, query_block: QueryBuilder):
        super().__init__(query_block)
        self.azimuth = _NumberQuery.init_with_limits("view:azimuth", query_block, min_value=0, max_value=360, is_int=False)
        self.incidence_angle = _NumberQuery.init_with_limits("view:incidence_angle", query_block, min_value=0, max_value=90, is_int=False)
        self.off_nadir = _NumberQuery.init_with_limits("view:off_nadir", query_block, min_value=0, max_value=90, is_int=False)
        self.sun_azimuth = _NumberQuery.init_with_limits("view:sun_azimuth", query_block, min_value=0, max_value=360, is_int=False)
        self.sun_elevation = _NumberQuery.init_with_limits("view:sun_elevation", query_block, min_value=-90, max_value=90, is_int=False)


class QueryBuilder:
    """
    class for building cql2-json queries

    ...

    Attributes
    ----------
    id : _StringQuery
        string query interface for identifier is unique within a Collection
    datetime : _DateQuery
        datetime query interface for searching the datetime of assets
    geometry : _SpatialQuery
        spatial query interface
    created : _DateQuery
        datetime query interface for searching items by the created field
    updated : _DateQuery
        datetime query interface for searching items by the updated field
    start_datetime : _DateQuery
        datetime query interface for searching items by the start_datetime field
    end_datetime : _DateQuery
        datetime query interface for searching items by the end_datetime field
    platform : _StringQuery
        string query interface for searching items by the platform field
    mission : _StringQuery
        string query interface for searching items by the mission field
    """

    def __init__(self):
        self._filter_expressions: list[_QueryTuple] = []
        self.id = _StringQuery("id", self)
        self.datetime = _DateQuery("datetime", self)
        self.geometry = _SpatialQuery("geometry", self)
        self.created = _DateQuery("created", self)
        self.updated = _DateQuery("updated", self)
        self.start_datetime = _DateQuery("start_datetime", self)
        self.end_datetime = _DateQuery("end_datetime", self)
        self.platform = _StringQuery("platform", self)
        self.mission = _StringQuery("mission", self)
        self.gsd = _NumberQuery.init_with_limits("gsd", self, min_value=0)
        self.eo = _EOExtension(self)
        self.landsat = _LandsatExtension(self)
        self.mlm = _MLMExtension(self)
        self.proj = _ProjExtension(self)
        self.sar = _SARExtension(self)
        self.sat = _SatExtension(self)
        self.view = _ViewExtension(self)

    def query_dump(self, top_level_is_or=False):
        properties = list(vars(self).values())
        args = [x._build_query() for x in properties if isinstance(x, _QueryBase) and x._build_query() is not None]
        for query_filter in self._filter_expressions:
            args.append(query_filter._build_query())

        for p in properties:
            if isinstance(p, _Extension):
                args.extend(p._build_query())

        if len(args) == 0:
            return None
        top_level_op = "and"
        if top_level_is_or:
            top_level_op = "or"
        return {
            "filter-lang": "cql2-json",
            "filter": {
                "op": top_level_op,
                "args": args}
        }

    def query_dump_json(self, top_level_is_or=False, indent=None, sort_keys=False):
        return json.dumps(self.query_dump(top_level_is_or=top_level_is_or),
                          indent=indent,
                          sort_keys=sort_keys,
                          cls=_DateTimeEncoder)

    def filter(self, *column_expression):
        query_tuple = column_expression[0]
        self._filter_expressions.append(query_tuple)


def filter_grouping(*column_expression):
    filter_tuple = _FilterTuple(column_expression[0].left, column_expression[0].op, column_expression[0].right)
    return filter_tuple
